@@ -5,6 +5,7 @@ import random
 import struct
 import sys
 
+from ctypes import create_string_buffer
 from puresasl import SASLError, SASLProtocolException
 from puresasl.util import bytes, num_to_bytes, bytes_to_num, quote
 
@@ -493,7 +494,34 @@ class GSSAPIMechanism(Mechanism):
         max_length, = struct.unpack('!i', '\x00' + plaintext_data[1:])
         self.max_buffer = min(self.sasl.max_buffer, max_length)
 
-        ret = kerberos.authGSSClientWrap(self.context, data, self.user)
+        """
+        Construct the reply.
+
+        byte 0: the selected qop. 1==auth, 2==auth-int, 4==auth-conf
+        byte 1-3: the max length for any buffer sent back and forth on
+            this connection. (big endian)
+        the rest of the buffer: the authorization user name in UTF-8 -
+            not null terminated.
+
+        So, we write the max length and authorization user name first, then
+        overwrite the first byte of the buffer with the qop.  This is ok since
+        the max length is writen out in big endian.
+        """
+        i = len(self.user)
+        fmt = '!I' + str(i) + 's'
+        outdata = create_string_buffer(4 + i)
+        struct.pack_into(fmt, outdata, 0, self.max_buffer, self.user)
+
+        qop = 1
+        if self.qop == 'auth-int':
+            qop = 2
+        elif self.qop == 'auth-conf':
+            qop = 4
+        struct.pack_into('!B', outdata, 0, qop)
+
+        encodeddata = base64.b64encode(outdata)
+
+        ret = kerberos.authGSSClientWrap(self.context, data)
         response = kerberos.authGSSClientResponse(self.context)
         self.complete = True
         return base64.b64decode(response)
@@ -501,8 +529,12 @@ class GSSAPIMechanism(Mechanism):
     def wrap(self, outgoing):
         if self.qop != 'auth':
             outgoing = base64.b64encode(outgoing)
-            kerberos.authGSSClientWrap(self.context, outgoing, self.user)
-            return base64.b64decode(kerberos.authGSSClientResponse, self.context)
+            if self.qop == 'auth-conf':
+                protect = 1
+            else:
+                protect = 0
+            kerberos.authGSSClientWrap(self.context, outgoing, None, protect)
+            return base64.b64decode(kerberos.authGSSClientResponse(self.context))
         else:
             return outgoing
 
@@ -510,7 +542,10 @@ class GSSAPIMechanism(Mechanism):
         if self.qop != 'auth':
             incoming = base64.b64encode(incoming)
             kerberos.authGSSClientUnwrap(self.context, incoming)
-            return base64.b64decode(kerberos.authGSSClientResponse, self.context)
+            conf = kerberos.authGSSClientResponseConf(self.context)
+            if 0 == conf and self.qop == 'auth-conf':
+                raise StandardError("Error: confidentiality requested, but not honored by the server.")
+            return base64.b64decode(kerberos.authGSSClientResponse(self.context))
         else:
             return incoming
 
